import abc
from os.path import abspath
from os.path import isfile
from os.path import normcase
from os.path import normpath
from os.path import sep
from os.path import splitdrive
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import attr
import six

from ddtrace.debugging._expressions import DDExpression
from ddtrace.internal.logger import get_logger
from ddtrace.internal.module import _resolve
from ddtrace.internal.rate_limiter import BudgetRateLimiterWithJitter as RateLimiter
from ddtrace.internal.utils.cache import cached


log = get_logger(__name__)

DEFAULT_PROBE_RATE = 5000.0
DEFAULT_SNAPSHOT_PROBE_RATE = 1.0
DEFAULT_PROBE_CONDITION_ERROR_RATE = 1.0 / 60 / 5


@cached()
def _resolve_source_file(path):
    # type: (str) -> Optional[str]
    """Resolve the source path for the given path.

    This recursively strips parent directories until it finds a file that
    exists according to sys.path.
    """
    npath = abspath(normpath(normcase(path)))
    if isfile(npath):
        return npath

    _, relpath = splitdrive(npath)
    while relpath:
        resolved_path = _resolve(relpath)
        if resolved_path is not None:
            return abspath(resolved_path)
        _, _, relpath = relpath.partition(sep)

    return None


MAXLEVEL = 2
MAXSIZE = 100
MAXLEN = 255
MAXFIELDS = 20


@attr.s
class CaptureLimits(object):
    max_level = attr.ib(type=int, default=MAXLEVEL)  # type: int
    max_size = attr.ib(type=int, default=MAXSIZE)  # type: int
    max_len = attr.ib(type=int, default=MAXLEN)  # type: int
    max_fields = attr.ib(type=int, default=MAXFIELDS)  # type: int


@attr.s(hash=True)
class Probe(six.with_metaclass(abc.ABCMeta)):
    probe_id = attr.ib(type=str)
    tags = attr.ib(type=dict, eq=False)
    rate = attr.ib(type=float, eq=False)
    limiter = attr.ib(type=RateLimiter, init=False, repr=False, eq=False)  # type: RateLimiter

    @limiter.default
    def _(self):
        return RateLimiter(
            limit_rate=self.rate,
            tau=1.0 / self.rate if self.rate else 1.0,
            on_exceed=lambda: log.warning("Rate limit exceeeded for %r", self),
            call_once=True,
            raise_on_exceed=False,
        )


@attr.s
class ProbeConditionMixin(object):
    """Conditional probe.

    If the condition is ``None``, then this is equivalent to a non-conditional
    probe.
    """

    condition = attr.ib(type=Optional[DDExpression])  # type: Optional[DDExpression]
    condition_error_rate = attr.ib(type=float, eq=False)
    condition_error_limiter = attr.ib(type=RateLimiter, init=False, repr=False, eq=False)  # type: RateLimiter

    @condition_error_limiter.default
    def _(self):
        return RateLimiter(
            limit_rate=self.condition_error_rate,
            tau=1.0 / self.condition_error_rate if self.condition_error_rate else 1.0,
            on_exceed=lambda: log.debug("Condition error rate limit exceeeded for %r", self),
            call_once=True,
            raise_on_exceed=False,
        )


@attr.s
class ProbeLocationMixin(object):
    def location(self):
        # type: () -> Tuple[str,str]
        """return a turple of (location,sublocation) for the probe.
        For example, line probe returns the (file,line) and method probe return (module,method)
        """
        return ("", "")


@attr.s
class LineLocationMixin(ProbeLocationMixin):
    source_file = attr.ib(type=str, converter=_resolve_source_file)  # type: ignore[misc]
    line = attr.ib(type=int)

    def location(self):
        return (self.source_file, self.line)


# TODO: make this an Enum once Python 2 support is dropped.
class ProbeEvaluateTimingForMethod(object):
    DEFAULT = "DEFAULT"
    ENTER = "ENTER"
    EXIT = "EXIT"


@attr.s
class FunctionLocationMixin(ProbeLocationMixin):
    module = attr.ib(type=str)
    func_qname = attr.ib(type=str)
    evaluate_at = attr.ib(type=ProbeEvaluateTimingForMethod)

    def location(self):
        return (self.module, self.func_qname)


# TODO: make this an Enum once Python 2 support is dropped.
class MetricProbeKind(object):
    COUNTER = "COUNT"
    GAUGE = "GAUGE"
    HISTOGRAM = "HISTOGRAM"
    DISTRIBUTION = "DISTRIBUTION"


@attr.s
class MetricProbeMixin(object):
    kind = attr.ib(type=str)
    name = attr.ib(type=str)
    value = attr.ib(type=Optional[Callable[[Dict[str, Any]], Any]])


@attr.s
class MetricLineProbe(Probe, LineLocationMixin, MetricProbeMixin, ProbeConditionMixin):
    pass


@attr.s
class MetricFunctionProbe(Probe, FunctionLocationMixin, MetricProbeMixin, ProbeConditionMixin):
    pass


@attr.s
class TemplateSegment(six.with_metaclass(abc.ABCMeta)):
    @abc.abstractmethod
    def eval(self, _locals):
        # type: (Dict[str,Any]) -> str
        pass


@attr.s
class LiteralTemplateSegment(TemplateSegment):
    str_value = attr.ib(type=str, default=None)

    def eval(self, _locals):
        # type: (Dict[str,Any]) -> Any
        return self.str_value


@attr.s
class ExpressionTemplateSegment(TemplateSegment):
    expr = attr.ib(type=DDExpression, default=None)  # type: DDExpression

    def eval(self, _locals):
        # type: (Dict[str,Any]) -> Any
        return self.expr.eval(_locals)


@attr.s
class LogProbeMixin(object):
    template = attr.ib(type=str)
    segments = attr.ib(type=List[TemplateSegment])
    take_snapshot = attr.ib(type=bool)
    limits = attr.ib(type=CaptureLimits, eq=False)  # type: CaptureLimits


@attr.s
class LogLineProbe(Probe, LineLocationMixin, LogProbeMixin, ProbeConditionMixin):
    pass


@attr.s
class LogFunctionProbe(Probe, FunctionLocationMixin, LogProbeMixin, ProbeConditionMixin):
    pass


LineProbe = Union[LogLineProbe, MetricLineProbe]
FunctionProbe = Union[LogFunctionProbe, MetricFunctionProbe]


class ProbeType(object):
    LOG_PROBE = "LOG_PROBE"
    METRIC_PROBE = "METRIC_PROBE"



"""
The Botocore integration will trace all AWS calls made with the botocore
library. Libraries like Boto3 that use Botocore will also be patched.

Enabling
~~~~~~~~

The botocore integration is enabled automatically when using
:ref:`ddtrace-run<ddtracerun>` or :func:`patch_all()<ddtrace.patch_all>`.

Or use :func:`patch()<ddtrace.patch>` to manually enable the integration::

    from ddtrace import patch
    patch(botocore=True)

Configuration
~~~~~~~~~~~~~

.. py:data:: ddtrace.config.botocore['distributed_tracing']

   Whether to inject distributed tracing data to requests in SQS, SNS, EventBridge, Kinesis Streams and Lambda.

   Can also be enabled with the ``DD_BOTOCORE_DISTRIBUTED_TRACING`` environment variable.

   Default: ``True``

.. py:data:: ddtrace.config.botocore['invoke_with_legacy_context']

    This preserves legacy behavior when tracing directly invoked Python and Node Lambda
    functions instrumented with datadog-lambda-python < v41 or datadog-lambda-js < v3.58.0.

    Legacy support for older libraries is available with
    ``ddtrace.config.botocore.invoke_with_legacy_context = True`` or by setting the environment
    variable ``DD_BOTOCORE_INVOKE_WITH_LEGACY_CONTEXT=true``.


    Default: ``False``


.. py:data:: ddtrace.config.botocore['operations'][<operation>].error_statuses = "<error statuses>"

    Definition of which HTTP status codes to consider for making a span as an error span.

    By default response status codes of ``'500-599'`` are considered as errors for all endpoints.

    Example marking 404, and 5xx as errors for ``s3.headobject`` API calls::

        from ddtrace import config

        config.botocore['operations']['s3.headobject'].error_statuses = '404,500-599'


    See :ref:`HTTP - Custom Error Codes<http-custom-error>` documentation for more examples.

.. py:data:: ddtrace.config.botocore['tag_no_params']

    This opts out of the default behavior of collecting a narrow set of API parameters as span tags.

    To not collect any API parameters, ``ddtrace.config.botocore.tag_no_params = True`` or by setting the environment
    variable ``DD_AWS_TAG_NO_PARAMS=true``.


    Default: ``False``

.. py:data:: ddtrace.config.botocore['tag_all_params']

    **Deprecated**: This retains the deprecated behavior of adding span tags for
    all API parameters that are not explicitly excluded by the integration.
    These deprecated span tags will be added along with the API parameters
    enabled by default.

    This configuration is ignored if ``tag_no_parms`` (``DD_AWS_TAG_NO_PARAMS``)
    is set to ``True``.

    To collect all API parameters, ``ddtrace.config.botocore.tag_all_params =
    True`` or by setting the environment variable ``DD_AWS_TAG_ALL_PARAMS=true``.


    Default: ``False``


Example::

    from ddtrace import config

    # Enable distributed tracing
    config.botocore['distributed_tracing'] = True

"""


from ...internal.utils.importlib import require_modules


required_modules = ["botocore.client"]

with require_modules(required_modules) as missing_modules:
    if not missing_modules:
        from .patch import patch

        __all__ = ["patch"]



from itertools import islice
from types import FrameType
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import Type

from ddtrace.debugging._probe.model import MAXFIELDS
from ddtrace.debugging._probe.model import MAXLEN
from ddtrace.debugging._probe.model import MAXLEVEL
from ddtrace.debugging._probe.model import MAXSIZE
from ddtrace.debugging.safety import get_fields
from ddtrace.internal.compat import BUILTIN_CONTAINER_TYPES
from ddtrace.internal.compat import BUILTIN_SIMPLE_TYPES
from ddtrace.internal.compat import CALLABLE_TYPES
from ddtrace.internal.compat import ExcInfoType
from ddtrace.internal.compat import NoneType
from ddtrace.internal.compat import stringify
from ddtrace.internal.safety import _isinstance
from ddtrace.internal.utils.cache import cached


if TYPE_CHECKING:  # pragma: no cover
    from ddtrace.internal.compat import Collection

EXCLUDED_FIELDS = frozenset(["__class__", "__dict__", "__weakref__", "__doc__", "__module__", "__hash__"])


@cached()
def qualname(_type):
    # type: (Type) -> str
    try:
        return stringify(_type.__qualname__)
    except AttributeError:
        # The logic for implementing qualname in Python 2 is complex, so if we
        # don't have it, we just return the name of the type.
        try:
            return _type.__name__
        except AttributeError:
            return repr(_type)


def _serialize_collection(value, brackets, level, maxsize, maxlen, maxfields):
    # type: (Collection, str, int, int, int, int) -> str
    o, c = brackets[0], brackets[1]
    ellipsis = ", ..." if len(value) > maxsize else ""
    return "".join(
        (o, ", ".join(serialize(_, level - 1, maxsize, maxlen, maxfields) for _ in islice(value, maxsize)), ellipsis, c)
    )


def serialize(value, level=MAXLEVEL, maxsize=MAXSIZE, maxlen=MAXLEN, maxfields=MAXFIELDS):
    # type: (Any, int, int, int, int) -> str
    """Python object serializer.

    We provide our own serializer to avoid any potential side effects of calling
    ``str`` directly on arbitrary objects.
    """

    if _isinstance(value, CALLABLE_TYPES):
        return object.__repr__(value)

    if type(value) in BUILTIN_SIMPLE_TYPES:
        r = repr(value)
        return "".join((r[:maxlen], "..." + ("'" if r[0] == "'" else "") if len(r) > maxlen else ""))

    if not level:
        return repr(type(value))

    if type(value) not in BUILTIN_CONTAINER_TYPES:
        return (
            type(value).__name__
            + "("
            + ", ".join(
                [
                    "=".join((k, serialize(v, level - 1, maxsize, maxlen, maxfields)))
                    for k, v in list(get_fields(value).items())[:maxfields]
                ]
            )
            + ")"
        )

    if type(value) is dict:
        return (
            "{"
            + ", ".join(
                [
                    ": ".join((serialize(k, level - 1, maxsize, maxlen, maxfields), serialize(v, level - 1)))
                    for k, v in value.items()
                ]
            )
            + "}"
        )
    elif type(value) is list:
        return _serialize_collection(value, "[]", level, maxsize, maxlen, maxfields)
    elif type(value) is tuple:
        return _serialize_collection(value, "()", level, maxsize, maxlen, maxfields)
    elif type(value) is set:
        return _serialize_collection(value, r"{}", level, maxsize, maxlen, maxfields) if value else "set()"

    raise TypeError("Unhandled type: %s", type(value))


def capture_stack(top_frame, max_height=4096):
    # type: (FrameType, int) -> List[dict]
    frame = top_frame  # type: Optional[FrameType]
    stack = []
    h = 0
    while frame and h < max_height:
        code = frame.f_code
        stack.append(
            {
                "fileName": code.co_filename,
                "function": code.co_name,
                "lineNumber": frame.f_lineno,
            }
        )
        frame = frame.f_back
        h += 1
    return stack


def capture_exc_info(exc_info):
    # type: (ExcInfoType) -> Optional[Dict[str, Any]]
    _type, value, tb = exc_info
    if _type is None or value is None:
        return None

    top_tb = tb
    if top_tb is not None:
        while top_tb.tb_next is not None:
            top_tb = top_tb.tb_next

    return {
        "type": _type.__name__,
        "message": ", ".join([serialize(v) for v in value.args]),
        "stacktrace": capture_stack(top_tb.tb_frame) if top_tb is not None else None,
    }


def capture_value(value, level=MAXLEVEL, maxlen=MAXLEN, maxsize=MAXSIZE, maxfields=MAXFIELDS):
    # type: (Any, int, int, int, int) -> Dict[str, Any]
    _type = type(value)

    if _type in BUILTIN_SIMPLE_TYPES:
        if _type is NoneType:
            return {"type": "NoneType", "isNull": True}

        value_repr = repr(value)
        value_repr_len = len(value_repr)
        return (
            {
                "type": qualname(_type),
                "value": value_repr,
            }
            if value_repr_len <= maxlen
            else {
                "type": qualname(_type),
                "value": value_repr[:maxlen],
                "truncated": True,
                "size": value_repr_len,
            }
        )

    if _type in BUILTIN_CONTAINER_TYPES:
        if level < 0:
            return {
                "type": qualname(_type),
                "notCapturedReason": "depth",
                "size": len(value),
            }

        if _type is dict:
            # Mapping
            data = {
                "type": "dict",
                "entries": [
                    (
                        capture_value(k, level=level - 1, maxlen=maxlen, maxsize=maxsize, maxfields=maxfields),
                        capture_value(v, level=level - 1, maxlen=maxlen, maxsize=maxsize, maxfields=maxfields),
                    )
                    for _, (k, v) in zip(range(maxsize), value.items())
                ],
                "size": len(value),
            }

        else:
            # Sequence
            data = {
                "type": qualname(_type),
                "elements": [
                    capture_value(v, level=level - 1, maxlen=maxlen, maxsize=maxsize, maxfields=maxfields)
                    for _, v in zip(range(maxsize), value)
                ],
                "size": len(value),
            }

        if len(value) > maxsize:
            data["notCapturedReason"] = "collectionSize"

        return data

    # Arbitrary object
    if level < 0:
        return {
            "type": qualname(_type),
            "notCapturedReason": "depth",
        }

    fields = get_fields(value)
    data = {
        "type": qualname(_type),
        "fields": {
            n: capture_value(v, level=level - 1, maxlen=maxlen, maxsize=maxsize, maxfields=maxfields)
            for _, (n, v) in zip(range(maxfields), fields.items())
        },
    }

    if len(fields) > maxfields:
        data["notCapturedReason"] = "fieldCount"

    return data



from ddtrace.debugging._expressions import DDExpression
from ddtrace.debugging._expressions import dd_compile
from ddtrace.debugging._probe.model import CaptureLimits
from ddtrace.debugging._probe.model import DEFAULT_PROBE_CONDITION_ERROR_RATE
from ddtrace.debugging._probe.model import DEFAULT_PROBE_RATE
from ddtrace.debugging._probe.model import DEFAULT_SNAPSHOT_PROBE_RATE
from ddtrace.debugging._probe.model import ExpressionTemplateSegment
from ddtrace.debugging._probe.model import LiteralTemplateSegment
from ddtrace.debugging._probe.model import LogFunctionProbe
from ddtrace.debugging._probe.model import LogLineProbe
from ddtrace.debugging._probe.model import MetricFunctionProbe
from ddtrace.debugging._probe.model import MetricLineProbe
from ddtrace.debugging._probe.model import ProbeEvaluateTimingForMethod


def compile_template(*args):
    template = ""
    segments = []

    for arg in args:
        if isinstance(arg, str):
            template += arg
            segments.append(LiteralTemplateSegment(arg))
        else:
            template += "{" + arg["dsl"] + "}"
            segments.append(ExpressionTemplateSegment(DDExpression(dsl=arg["dsl"], callable=dd_compile(arg["json"]))))

    return {"template": template, "segments": segments}


def create_probe_defaults(f):
    def _wrapper(*args, **kwargs):
        kwargs.setdefault("tags", dict())
        return f(*args, **kwargs)

    return _wrapper


def probe_conditional_defaults(f):
    def _wrapper(*args, **kwargs):
        kwargs.setdefault("condition", None)
        kwargs.setdefault("condition_error_rate", DEFAULT_PROBE_CONDITION_ERROR_RATE)
        return f(*args, **kwargs)

    return _wrapper


def function_location_defaults(f):
    def _wrapper(*args, **kwargs):
        kwargs.setdefault("evaluate_at", ProbeEvaluateTimingForMethod.DEFAULT)
        return f(*args, **kwargs)

    return _wrapper


def log_probe_defaults(f):
    def _wrapper(*args, **kwargs):
        kwargs.setdefault("take_snapshot", False)
        kwargs.setdefault("rate", DEFAULT_PROBE_RATE)
        kwargs.setdefault("limits", CaptureLimits())
        return f(*args, **kwargs)

    return _wrapper


def snapshot_probe_defaults(f):
    def _wrapper(*args, **kwargs):
        kwargs.setdefault("take_snapshot", True)
        kwargs.setdefault("rate", DEFAULT_SNAPSHOT_PROBE_RATE)
        kwargs.setdefault("limits", CaptureLimits())
        kwargs.setdefault("template", "")
        kwargs.setdefault("segments", [])
        return f(*args, **kwargs)

    return _wrapper


def metric_probe_defaults(f):
    def _wrapper(*args, **kwargs):
        kwargs.setdefault("rate", DEFAULT_PROBE_RATE)
        kwargs.setdefault("value", None)
        return f(*args, **kwargs)

    return _wrapper


@create_probe_defaults
@probe_conditional_defaults
@snapshot_probe_defaults
def create_snapshot_line_probe(**kwargs):
    return LogLineProbe(**kwargs)


@create_probe_defaults
@probe_conditional_defaults
@function_location_defaults
@snapshot_probe_defaults
def create_snapshot_function_probe(**kwargs):
    return LogFunctionProbe(**kwargs)


@create_probe_defaults
@probe_conditional_defaults
@log_probe_defaults
def create_log_line_probe(**kwargs):
    return LogLineProbe(**kwargs)


@create_probe_defaults
@probe_conditional_defaults
@function_location_defaults
@log_probe_defaults
def create_log_function_probe(**kwargs):
    return LogFunctionProbe(**kwargs)


@create_probe_defaults
@probe_conditional_defaults
@metric_probe_defaults
def create_metric_line_probe(**kwargs):
    return MetricLineProbe(**kwargs)


@create_probe_defaults
@probe_conditional_defaults
@function_location_defaults
@metric_probe_defaults
def create_metric_function_probe(**kwargs):
    return MetricFunctionProbe(**kwargs)



import sys


sys.path.insert(0, "../../..")  # to allow importing from the tests module.

from tests.debugging.mocking import TestDebugger  # noqa
from tests.debugging.utils import create_snapshot_line_probe  # noqa


TestDebugger.enable()

debugger = TestDebugger._instance

debugger.add_probes(
    create_snapshot_line_probe(
        probe_id="run_module_test",
        source_file="target.py",
        line=9,
        condition=None,
    )
)



from contextlib import contextmanager
import os
import subprocess

import django
import pytest

from ddtrace.internal.compat import PY3
import ddtrace.internal.constants as constants
from tests.appsec.test_processor import RULES_GOOD_PATH
from tests.appsec.test_processor import _ALLOWED_IP
from tests.appsec.test_processor import _BLOCKED_IP
from tests.utils import snapshot
from tests.webclient import Client


SERVER_PORT = 8000


@contextmanager
def daphne_client(django_asgi, additional_env=None):
    """Runs a django app hosted with a daphne webserver in a subprocess and
    returns a client which can be used to query it.

    Traces are flushed by invoking a tracer.shutdown() using a /shutdown-tracer route
    at the end of the testcase.
    """

    # Make sure to copy the environment as we need the PYTHONPATH and _DD_TRACE_WRITER_ADDITIONAL_HEADERS (for the test
    # token) propagated to the new process.
    env = os.environ.copy()
    env.update(additional_env or {})
    assert "_DD_TRACE_WRITER_ADDITIONAL_HEADERS" in env, "Client fixture needs test token in headers"
    env.update(
        {
            "DJANGO_SETTINGS_MODULE": "tests.contrib.django.django_app.settings",
        }
    )

    # ddtrace-run uses execl which replaces the process but the webserver process itself might spawn new processes.
    # Right now it doesn't but it's possible that it might in the future (ex. uwsgi).
    cmd = ["ddtrace-run", "daphne", "-p", str(SERVER_PORT), "tests.contrib.django.asgi:%s" % django_asgi]
    proc = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        close_fds=True,
        env=env,
    )

    client = Client("http://localhost:%d" % SERVER_PORT)

    # Wait for the server to start up
    client.wait()

    try:
        yield client
    finally:
        resp = client.get_ignored("/shutdown-tracer")
        assert resp.status_code == 200
        proc.terminate()


@pytest.mark.skipif(django.VERSION < (3, 2, 0), reason="Only want to test with latest Django")
@snapshot(
    ignores=[
        "meta.error.stack",
        "meta.http.request.headers.user-agent",
        "meta.http.useragent",
        "metrics._dd.appsec.waf.duration",
        "metrics._dd.appsec.waf.duration_ext",
    ]
)
def test_appsec_enabled():
    with daphne_client("application", additional_env={"DD_APPSEC_ENABLED": "true"}) as client:
        resp = client.get("/")
        assert resp.status_code == 200
        assert resp.content == b"Hello, test app."


@pytest.mark.skipif(django.VERSION < (3, 2, 0), reason="Only want to test with latest Django")
@snapshot(
    ignores=[
        "meta.error.stack",
        "meta.http.request.headers.user-agent",
        "meta.http.useragent",
        "metrics._dd.appsec.waf.duration",
        "metrics._dd.appsec.waf.duration_ext",
    ]
)
def test_appsec_enabled_attack():
    with daphne_client("application", additional_env={"DD_APPSEC_ENABLED": "true"}) as client:
        resp = client.get("/.git")
        assert resp.status_code == 404


@pytest.mark.skipif(django.VERSION < (3, 2, 0), reason="Only want to test with latest Django")
@snapshot(
    ignores=[
        "meta.http.request.headers.accept-encoding",
        "meta.http.request.headers.user-agent",
        "meta.http.useragent",
        "metrics._dd.appsec.waf.duration",
        "metrics._dd.appsec.waf.duration_ext",
        "metrics._dd.appsec.event_rules.loaded",
    ]
)
def test_request_ipblock_nomatch_200():
    with daphne_client(
        "application",
        additional_env={
            "DD_DJANGO_INSTRUMENT_TEMPLATES": "false",
            "DD_APPSEC_ENABLED": "true",
            "DD_APPSEC_RULES": RULES_GOOD_PATH,
        },
    ) as client:
        result = client.get("/", headers={"X-Real-Ip": _ALLOWED_IP})
        assert result.status_code == 200
        assert result.content == b"Hello, test app."


@pytest.mark.skipif(django.VERSION < (3, 2, 0), reason="Only want to test with latest Django")
@snapshot(
    ignores=[
        "meta._dd.appsec.waf.duration",
        "meta._dd.appsec.waf.duration_ext",
        "meta.http.request.headers.accept-encoding",
        "meta.http.request.headers.user-agent",
        "meta.http.useragent",
        "metrics._dd.appsec.waf.duration",
        "metrics._dd.appsec.waf.duration_ext",
        "metrics._dd.appsec.event_rules.loaded",
    ]
)
def test_request_ipblock_match_403():
    with daphne_client(
        "application",
        additional_env={
            "DD_APPSEC_ENABLED": "true",
            "DD_APPSEC_RULES": RULES_GOOD_PATH,
        },
    ) as client:
        result = client.get(
            "/",
            headers={
                "X-Real-Ip": _BLOCKED_IP,
                "Accept": "text/html",
            },
        )
        assert result.status_code == 403
        as_bytes = (
            bytes(constants.APPSEC_BLOCKED_RESPONSE_HTML, "utf-8") if PY3 else constants.APPSEC_BLOCKED_RESPONSE_HTML
        )
        assert result.content == as_bytes


@pytest.mark.skipif(django.VERSION < (3, 2, 0), reason="Only want to test with latest Django")
@snapshot(
    ignores=[
        "meta._dd.appsec.waf.duration",
        "meta._dd.appsec.waf.duration_ext",
        "meta.http.request.headers.accept-encoding",
        "meta.http.request.headers.user-agent",
        "meta.http.useragent",
        "metrics._dd.appsec.waf.duration",
        "metrics._dd.appsec.waf.duration_ext",
        "metrics._dd.appsec.event_rules.loaded",
    ]
)
def test_request_ipblock_match_403_json():
    with daphne_client(
        "application",
        additional_env={
            "DD_APPSEC_ENABLED": "true",
            "DD_APPSEC_RULES": RULES_GOOD_PATH,
        },
    ) as client:
        result = client.get(
            "/",
            headers={
                "X-Real-Ip": _BLOCKED_IP,
            },
        )
        assert result.status_code == 403
        as_bytes = (
            bytes(constants.APPSEC_BLOCKED_RESPONSE_JSON, "utf-8") if PY3 else constants.APPSEC_BLOCKED_RESPONSE_JSONP
        )
        assert result.content == as_bytes



TELEMETRY_METRIC_TYPE_COUNT = "count"
TELEMETRY_METRIC_TYPE_GAUGE = "gauge"
TELEMETRY_METRIC_TYPE_RATE = "rate"
TELEMETRY_METRIC_TYPE_DISTRIBUTIONS = "distributions"

TELEMETRY_NAMESPACE_TAG_TRACER = "tracers"
TELEMETRY_NAMESPACE_TAG_APPSEC = "appsec"

TELEMETRY_TYPE_GENERATE_METRICS = "generate-metrics"
TELEMETRY_TYPE_DISTRIBUTION = "distributions"



# stdlib
import contextlib
import json
from typing import Iterable

# 3p
import pymongo

# project
import ddtrace
from ddtrace import config
from ddtrace.internal.constants import COMPONENT
from ddtrace.vendor.wrapt import ObjectProxy

from ...constants import ANALYTICS_SAMPLE_RATE_KEY
from ...constants import SPAN_KIND
from ...constants import SPAN_MEASURED_KEY
from ...ext import SpanKind
from ...ext import SpanTypes
from ...ext import db
from ...ext import mongo as mongox
from ...ext import net as netx
from ...internal.compat import iteritems
from ...internal.logger import get_logger
from ...internal.utils import get_argument_value
from .parse import parse_msg
from .parse import parse_query
from .parse import parse_spec


BATCH_PARTIAL_KEY = "Batch"

# Original Client class
_MongoClient = pymongo.MongoClient

VERSION = pymongo.version_tuple

if VERSION < (3, 6, 0):
    from pymongo.helpers import _unpack_response


log = get_logger(__name__)


class TracedMongoClient(ObjectProxy):
    def __init__(self, client=None, *args, **kwargs):
        # To support the former trace_mongo_client interface, we have to keep this old interface
        # TODO(Benjamin): drop it in a later version
        if not isinstance(client, _MongoClient):
            # Patched interface, instantiate the client

            # client is just the first arg which could be the host if it is
            # None, then it could be that the caller:

            # if client is None then __init__ was:
            #   1) invoked with host=None
            #   2) not given a first argument (client defaults to None)
            # we cannot tell which case it is, but it should not matter since
            # the default value for host is None, in either case we can simply
            # not provide it as an argument
            if client is None:
                client = _MongoClient(*args, **kwargs)
            # else client is a value for host so just pass it along
            else:
                client = _MongoClient(client, *args, **kwargs)

        super(TracedMongoClient, self).__init__(client)
        # NOTE[matt] the TracedMongoClient attempts to trace all of the network
        # calls in the trace library. This is good because it measures the
        # actual network time. It's bad because it uses a private API which
        # could change. We'll see how this goes.
        if not isinstance(client._topology, TracedTopology):
            client._topology = TracedTopology(client._topology)

        # Default Pin
        ddtrace.Pin(service="pymongo").onto(self)

    def __setddpin__(self, pin):
        pin.onto(self._topology)

    def __getddpin__(self):
        return ddtrace.Pin.get_from(self._topology)


class TracedTopology(ObjectProxy):
    def __init__(self, topology):
        super(TracedTopology, self).__init__(topology)

    def select_server(self, *args, **kwargs):
        s = self.__wrapped__.select_server(*args, **kwargs)
        if not isinstance(s, TracedServer):
            s = TracedServer(s)
        # Reattach the pin every time in case it changed since the initial patching
        ddtrace.Pin.get_from(self).onto(s)
        return s


class TracedServer(ObjectProxy):
    def __init__(self, server):
        super(TracedServer, self).__init__(server)

    def _datadog_trace_operation(self, operation):
        cmd = None
        # Only try to parse something we think is a query.
        if self._is_query(operation):
            try:
                cmd = parse_query(operation)
            except Exception:
                log.exception("error parsing query")

        pin = ddtrace.Pin.get_from(self)
        # if we couldn't parse or shouldn't trace the message, just go.
        if not cmd or not pin or not pin.enabled():
            return None

        span = pin.tracer.trace("pymongo.cmd", span_type=SpanTypes.MONGODB, service=pin.service)

        span.set_tag_str(COMPONENT, config.pymongo.integration_name)

        # set span.kind to the operation type being performed
        span.set_tag_str(SPAN_KIND, SpanKind.CLIENT)

        span.set_tag(SPAN_MEASURED_KEY)
        span.set_tag_str(mongox.DB, cmd.db)
        span.set_tag_str(mongox.COLLECTION, cmd.coll)
        span.set_tag_str(db.SYSTEM, mongox.SERVICE)
        span.set_tags(cmd.tags)

        # set `mongodb.query` tag and resource for span
        _set_query_metadata(span, cmd)

        # set analytics sample rate
        sample_rate = config.pymongo.get_analytics_sample_rate()
        if sample_rate is not None:
            span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, sample_rate)
        return span

    # Pymongo >= 3.12
    if VERSION >= (3, 12, 0):

        def run_operation(self, sock_info, operation, *args, **kwargs):
            span = self._datadog_trace_operation(operation)
            if span is None:
                return self.__wrapped__.run_operation(sock_info, operation, *args, **kwargs)
            with span:
                result = self.__wrapped__.run_operation(sock_info, operation, *args, **kwargs)
                if result:
                    if hasattr(result, "address"):
                        set_address_tags(span, result.address)
                    if self._is_query(operation) and hasattr(result, "docs"):
                        set_query_rowcount(docs=result.docs, span=span)
                return result

    # Pymongo >= 3.9, <3.12
    elif (3, 9, 0) <= VERSION < (3, 12, 0):

        def run_operation_with_response(self, sock_info, operation, *args, **kwargs):
            span = self._datadog_trace_operation(operation)
            if span is None:
                return self.__wrapped__.run_operation_with_response(sock_info, operation, *args, **kwargs)
            with span:
                result = self.__wrapped__.run_operation_with_response(sock_info, operation, *args, **kwargs)
                if result:
                    if hasattr(result, "address"):
                        set_address_tags(span, result.address)
                    if self._is_query(operation) and hasattr(result, "docs"):
                        set_query_rowcount(docs=result.docs, span=span)
                return result

    # Pymongo < 3.9
    else:

        def send_message_with_response(self, operation, *args, **kwargs):
            span = self._datadog_trace_operation(operation)
            if span is None:
                return self.__wrapped__.send_message_with_response(operation, *args, **kwargs)
            with span:
                result = self.__wrapped__.send_message_with_response(operation, *args, **kwargs)
                if result:
                    if hasattr(result, "address"):
                        set_address_tags(span, result.address)
                    if self._is_query(operation):
                        if hasattr(result, "data"):
                            if VERSION >= (3, 6, 0) and hasattr(result.data, "unpack_response"):
                                set_query_rowcount(docs=result.data.unpack_response(), span=span)
                            else:
                                data = _unpack_response(response=result.data)
                                if VERSION < (3, 2, 0) and data.get("number_returned", None):
                                    span.set_metric(db.ROWCOUNT, data.get("number_returned"))
                                elif (3, 2, 0) <= VERSION < (3, 6, 0):
                                    docs = data.get("data", None)
                                    set_query_rowcount(docs=docs, span=span)
                return result

    @contextlib.contextmanager
    def get_socket(self, *args, **kwargs):
        with self.__wrapped__.get_socket(*args, **kwargs) as s:
            if not isinstance(s, TracedSocket):
                s = TracedSocket(s)
            ddtrace.Pin.get_from(self).onto(s)
            yield s

    @staticmethod
    def _is_query(op):
        # NOTE: _Query should always have a spec field
        return hasattr(op, "spec")


class TracedSocket(ObjectProxy):
    def __init__(self, socket):
        super(TracedSocket, self).__init__(socket)

    def command(self, dbname, spec, *args, **kwargs):
        cmd = None
        try:
            cmd = parse_spec(spec, dbname)
        except Exception:
            log.exception("error parsing spec. skipping trace")

        pin = ddtrace.Pin.get_from(self)
        # skip tracing if we don't have a piece of data we need
        if not dbname or not cmd or not pin or not pin.enabled():
            return self.__wrapped__.command(dbname, spec, *args, **kwargs)

        cmd.db = dbname
        with self.__trace(cmd):
            return self.__wrapped__.command(dbname, spec, *args, **kwargs)

    def write_command(self, *args, **kwargs):
        msg = get_argument_value(args, kwargs, 1, "msg")
        cmd = None
        try:
            cmd = parse_msg(msg)
        except Exception:
            log.exception("error parsing msg")

        pin = ddtrace.Pin.get_from(self)
        # if we couldn't parse it, don't try to trace it.
        if not cmd or not pin or not pin.enabled():
            return self.__wrapped__.write_command(*args, **kwargs)

        with self.__trace(cmd) as s:
            result = self.__wrapped__.write_command(*args, **kwargs)
            if result:
                s.set_metric(db.ROWCOUNT, result.get("n", -1))
            return result

    def __trace(self, cmd):
        pin = ddtrace.Pin.get_from(self)
        s = pin.tracer.trace("pymongo.cmd", span_type=SpanTypes.MONGODB, service=pin.service)

        s.set_tag_str(COMPONENT, config.pymongo.integration_name)
        s.set_tag_str(db.SYSTEM, mongox.SERVICE)

        # set span.kind to the type of operation being performed
        s.set_tag_str(SPAN_KIND, SpanKind.CLIENT)

        s.set_tag(SPAN_MEASURED_KEY)
        if cmd.db:
            s.set_tag_str(mongox.DB, cmd.db)
        if cmd:
            s.set_tag(mongox.COLLECTION, cmd.coll)
            s.set_tags(cmd.tags)
            s.set_metrics(cmd.metrics)

        # set `mongodb.query` tag and resource for span
        _set_query_metadata(s, cmd)

        # set analytics sample rate
        s.set_tag(ANALYTICS_SAMPLE_RATE_KEY, config.pymongo.get_analytics_sample_rate())

        if self.address:
            set_address_tags(s, self.address)
        return s


def normalize_filter(f=None):
    if f is None:
        return {}
    elif isinstance(f, list):
        # normalize lists of filters
        # e.g. {$or: [ { age: { $lt: 30 } }, { type: 1 } ]}
        return [normalize_filter(s) for s in f]
    elif isinstance(f, dict):
        # normalize dicts of filters
        #   {$or: [ { age: { $lt: 30 } }, { type: 1 } ]})
        out = {}
        for k, v in iteritems(f):
            if k == "$in" or k == "$nin":
                # special case $in queries so we don't loop over lists.
                out[k] = "?"
            elif isinstance(v, list) or isinstance(v, dict):
                # RECURSION ALERT: needs to move to the agent
                out[k] = normalize_filter(v)
            else:
                # NOTE: this shouldn't happen, but let's have a safeguard.
                out[k] = "?"
        return out
    else:
        # FIXME[matt] unexpected type. not sure this should ever happen, but at
        # least it won't crash.
        return {}


def set_address_tags(span, address):
    # the address is only set after the cursor is done.
    if address:
        span.set_tag_str(netx.TARGET_HOST, address[0])
        span.set_tag(netx.TARGET_PORT, address[1])


def _set_query_metadata(span, cmd):
    """Sets span `mongodb.query` tag and resource given command query"""
    if cmd.query:
        nq = normalize_filter(cmd.query)
        span.set_tag("mongodb.query", nq)
        # needed to dump json so we don't get unicode
        # dict keys like {u'foo':'bar'}
        q = json.dumps(nq)
        span.resource = "{} {} {}".format(cmd.name, cmd.coll, q)
    else:
        span.resource = "{} {}".format(cmd.name, cmd.coll)


def set_query_rowcount(docs, span):
    # results returned in batches, get len of each batch
    if isinstance(docs, Iterable) and len(docs) > 0:
        cursor = docs[0].get("cursor", None)
    if cursor:
        rowcount = sum([len(documents) for batch_key, documents in cursor.items() if BATCH_PARTIAL_KEY in batch_key])
        span.set_metric(db.ROWCOUNT, rowcount)



import os

import pytest

import ddtrace.appsec.utils as utils
from ddtrace.internal.constants import APPSEC_BLOCKED_RESPONSE_HTML
from ddtrace.internal.constants import APPSEC_BLOCKED_RESPONSE_JSON
from tests.utils import override_env


@pytest.fixture(autouse=True)
def reset_template_caches():
    utils._HTML_BLOCKED_TEMPLATE_CACHE = None
    utils._JSON_BLOCKED_TEMPLATE_CACHE = None
    try:
        yield
    finally:
        utils._HTML_BLOCKED_TEMPLATE_CACHE = None
        utils._JSON_BLOCKED_TEMPLATE_CACHE = None


def test_get_blocked_template_no_env_var_html():
    with override_env(dict(DD_APPSEC_HTTP_BLOCKED_TEMPLATE_HTML="", DD_APPSEC_HTTP_BLOCKED_TEMPLATE_JSON="")):
        assert utils._get_blocked_template("text/html") == APPSEC_BLOCKED_RESPONSE_HTML


def test_get_blocked_template_no_env_var_json():
    with override_env(dict(DD_APPSEC_HTTP_BLOCKED_TEMPLATE_HTML="", DD_APPSEC_HTTP_BLOCKED_TEMPLATE_JSON="")):
        assert utils._get_blocked_template("other") == APPSEC_BLOCKED_RESPONSE_JSON
        assert utils._get_blocked_template("application/json") == APPSEC_BLOCKED_RESPONSE_JSON
        assert utils._get_blocked_template("") == APPSEC_BLOCKED_RESPONSE_JSON
        assert utils._get_blocked_template(None) == APPSEC_BLOCKED_RESPONSE_JSON


def test_get_blocked_template_user_file_missing_html():
    with override_env(dict(DD_APPSEC_HTTP_BLOCKED_TEMPLATE_HTML="missing.html")):
        assert utils._get_blocked_template("text/html") == APPSEC_BLOCKED_RESPONSE_HTML
        assert utils._get_blocked_template("") == APPSEC_BLOCKED_RESPONSE_JSON
        assert utils._get_blocked_template("application/json") == APPSEC_BLOCKED_RESPONSE_JSON


def test_get_blocked_template_user_file_missing_json():
    with override_env(dict(DD_APPSEC_HTTP_BLOCKED_TEMPLATE_JSON="missing.json")):
        assert utils._get_blocked_template("") == APPSEC_BLOCKED_RESPONSE_JSON
        assert utils._get_blocked_template("application/json") == APPSEC_BLOCKED_RESPONSE_JSON
        assert utils._get_blocked_template("text/html") == APPSEC_BLOCKED_RESPONSE_HTML


def test_get_blocked_template_user_file_exists_html():
    template_path = os.path.join(os.path.dirname(__file__), "blocking_template_html.html")
    with override_env(dict(DD_APPSEC_HTTP_BLOCKED_TEMPLATE_HTML=template_path)):
        with open(template_path, "r") as test_template_html:
            html_content = test_template_html.read()
        assert utils._get_blocked_template("text/html") == html_content
        assert utils._get_blocked_template("") == APPSEC_BLOCKED_RESPONSE_JSON
        assert utils._get_blocked_template("application/json") == APPSEC_BLOCKED_RESPONSE_JSON


def test_get_blocked_template_user_file_exists_json():
    template_path = os.path.join(os.path.dirname(__file__), "blocking_template_json.json")
    with override_env(dict(DD_APPSEC_HTTP_BLOCKED_TEMPLATE_JSON=template_path)):
        with open(template_path, "r") as test_template_json:
            json_content = test_template_json.read()
        assert utils._get_blocked_template("") == json_content
        assert utils._get_blocked_template("application/json") == json_content
        assert utils._get_blocked_template("text/html") == APPSEC_BLOCKED_RESPONSE_HTML



import ctypes
import time
from typing import TYPE_CHECKING

from six import text_type

from ddtrace.internal.logger import get_logger


if TYPE_CHECKING:
    from typing import Any
    from typing import Union

    from ddtrace.appsec.ddwaf.ddwaf_types import DDWafRulesType


LOGGER = get_logger(__name__)

try:
    from .ddwaf_types import ddwaf_config
    from .ddwaf_types import ddwaf_context_capsule
    from .ddwaf_types import ddwaf_get_version
    from .ddwaf_types import ddwaf_object
    from .ddwaf_types import ddwaf_object_free
    from .ddwaf_types import ddwaf_result
    from .ddwaf_types import ddwaf_ruleset_info
    from .ddwaf_types import ddwaf_run
    from .ddwaf_types import py_ddwaf_context_init
    from .ddwaf_types import py_ddwaf_init
    from .ddwaf_types import py_ddwaf_required_addresses
    from .ddwaf_types import py_ddwaf_update

    _DDWAF_LOADED = True
except OSError:
    _DDWAF_LOADED = False
    LOGGER.warning("DDWaf features disabled. WARNING: Dynamic Library not loaded", exc_info=True)

#
# Interface as Cython
#

DEFAULT_DDWAF_TIMEOUT_MS = 2


class DDWaf_result(object):
    __slots__ = ["data", "actions", "runtime", "total_runtime"]

    def __init__(self, data, actions, runtime, total_runtime):
        # type: (DDWaf_result, text_type|None, list[text_type], float, float) -> None
        self.data = data
        self.actions = actions
        self.runtime = runtime
        self.total_runtime = total_runtime


class DDWaf_info(object):
    __slots__ = ["loaded", "failed", "errors", "version"]

    def __init__(self, loaded, failed, errors, version):
        # type: (DDWaf_info, int, int, dict[text_type, Any], text_type) -> None
        self.loaded = loaded
        self.failed = failed
        self.errors = errors
        self.version = version

    def __repr__(self):
        return "{loaded: %d, failed: %d, errors: %s, version: %s}" % (
            self.loaded,
            self.failed,
            str(self.errors),
            self.version,
        )


if _DDWAF_LOADED:

    class DDWaf(object):
        def __init__(self, ruleset_map, obfuscation_parameter_key_regexp, obfuscation_parameter_value_regexp):
            # type: (DDWaf, dict[text_type, Any], text_type, text_type) -> None
            config = ddwaf_config(
                key_regex=obfuscation_parameter_key_regexp, value_regex=obfuscation_parameter_value_regexp
            )
            self._info = ddwaf_ruleset_info()
            ruleset_map_object = ddwaf_object.create_without_limits(ruleset_map)
            self._handle = py_ddwaf_init(ruleset_map_object, ctypes.byref(config), ctypes.byref(self._info))
            if not self._handle or self._info.failed:
                # We keep the handle alive in case of errors, as some valid rules can be loaded
                # at the same time some invalid ones are rejected
                LOGGER.error(
                    "DDWAF.__init__: invalid rules\n ruleset: %s\nloaded:%s\nerrors:%s\n",
                    ruleset_map_object.struct,
                    self._info.loaded,
                    self.info.errors,
                )
            ddwaf_object_free(ctypes.byref(ruleset_map_object))

        @property
        def required_data(self):
            # type: (DDWaf) -> list[text_type]
            return py_ddwaf_required_addresses(self._handle) if self._handle else []

        @property
        def info(self):
            # type: (DDWaf) -> DDWaf_info
            errors_result = self._info.errors.struct if self._info.failed > 0 else {}
            version = self._info.version
            version = "" if version is None else version.decode("UTF-8")
            return DDWaf_info(self._info.loaded, self._info.failed, errors_result, version)

        def update_rules(self, new_rules):
            # type: (dict[text_type, DDWafRulesType]) -> bool
            """update the rules of the WAF instance. return True if an error occurs."""
            rules = ddwaf_object.create_without_limits(new_rules)
            result = py_ddwaf_update(self._handle, rules, self._info)
            ddwaf_object_free(rules)
            if result:
                LOGGER.debug("DDWAF.update_rules success.\ninfo %s", self.info)
                self._handle = result
                return True
            else:
                LOGGER.debug("DDWAF.update_rules: keeping the previous handle.")
                return False

        def _at_request_start(self):
            # type: () -> ddwaf_context_capsule
            if self._handle:
                ctx = py_ddwaf_context_init(self._handle)
            if not ctx:
                LOGGER.error("DDWaf._at_request_start: failure to create the context.")
            return ctx

        def _at_request_end(self):
            # () -> None
            pass

        def run(
            self,  # type: DDWaf
            ctx,  # type: ddwaf_context_capsule
            data,  # type: DDWafRulesType
            timeout_ms=DEFAULT_DDWAF_TIMEOUT_MS,  # type:int
        ):
            # type: (...) -> DDWaf_result
            start = time.time()

            if not ctx:
                LOGGER.error("DDWaf.run: dry run. no context created.")
                return DDWaf_result(None, [], 0, (time.time() - start) * 1e6)

            result = ddwaf_result()
            wrapper = ddwaf_object(data)
            error = ddwaf_run(ctx.ctx, wrapper, ctypes.byref(result), timeout_ms * 1000)
            if error < 0:
                LOGGER.warning("run DDWAF error: %d\ninput %s\nerror %s", error, wrapper.struct, self.info.errors)
            return DDWaf_result(
                result.data.decode("UTF-8", errors="ignore") if hasattr(result, "data") and result.data else None,
                [result.actions.array[i].decode("UTF-8", errors="ignore") for i in range(result.actions.size)],
                result.total_runtime / 1e3,
                (time.time() - start) * 1e6,
            )

    def version():
        # type: () -> text_type
        return ddwaf_get_version().decode("UTF-8")


else:
    # Mockup of the DDWaf class doing nothing
    class DDWaf(object):  # type: ignore
        required_data = []  # type: list[text_type]
        info = DDWaf_info(0, 0, {}, "")  # type: DDWaf_info

        def __init__(self, rules, obfuscation_parameter_key_regexp, obfuscation_parameter_value_regexp):
            # type: (DDWaf, Union[None, int, text_type, list[Any], dict[text_type, Any]], text_type, text_type) -> None
            pass

        def run(
            self,  # type: DDWaf
            ctx,  # type: ddwaf_context_capsule
            data,  # type: Union[None, int, text_type, list[Any], dict[text_type, Any]]
            timeout_ms=DEFAULT_DDWAF_TIMEOUT_MS,  # type:int
        ):
            # type: (...) -> DDWaf_result
            LOGGER.warning("DDWaf features disabled. dry run")
            return DDWaf_result(None, [], 0.0, 0.0)

        def update_rules(self, _):
            # type: (dict[text_type, DDWafRulesType]) -> bool
            LOGGER.warning("DDWaf features disabled. dry update")
            return False

        def _at_request_start(self):
            pass

        def _at_request_end(self):
            pass

    def version():
        # type: () -> text_type
        LOGGER.warning("DDWaf features disabled. null version")
        return "0.0.0"



import inspect
import sys

import pytest

from ddtrace.internal.wrapping import wrap


@pytest.mark.asyncio
async def test_async_generator():
    async def stream():
        yield b"hello"
        yield b""
        return

    async def body():
        chunks = []
        async for chunk in stream():
            chunks.append(chunk)
        _body = b"".join(chunks)
        return _body

    wrapper_called = awrapper_called = False

    async def wrapper(f, args, kwargs):
        nonlocal wrapper_called
        wrapper_called = True
        return await f(*args, **kwargs)

    async def agwrapper(f, args, kwargs):
        nonlocal awrapper_called
        awrapper_called = True
        async for _ in f(*args, **kwargs):
            yield _

    wrap(stream, agwrapper)
    wrap(body, wrapper)

    assert await body() == b"hello"
    assert wrapper_called
    assert awrapper_called


@pytest.mark.asyncio
async def test_wrap_async_generator_send():
    def wrapper(f, args, kwargs):
        return f(*args, **kwargs)

    async def g():
        yield 0
        for _ in range(1, 10):
            n = yield _
            assert _ == n
        return

    wrap(g, wrapper)

    channel = []

    async def consume():
        agen = g()
        n = await agen.__anext__()
        channel.append(n)
        try:
            while True:
                n = await agen.asend(n)
                channel.append(n)
        except StopAsyncIteration:
            pass

        assert list(range(10)) == channel

    await consume()


@pytest.mark.asyncio
async def test_double_async_for_with_exception():
    channel = None

    class StreamConsumed(Exception):
        pass

    class AsyncIteratorByteStream(object):
        def __init__(self, stream):
            self._stream = stream
            self._is_stream_consumed = False

        async def __aiter__(self):
            if self._is_stream_consumed:
                raise StreamConsumed()

            self._is_stream_consumed = True
            async for part in self._stream:
                yield part

    async def wrapper(f, args, kwargs):
        nonlocal channel

        channel = [_ async for _ in f(*args, **kwargs)]
        for _ in channel:
            yield _
        return

    async def stream():
        yield b"hello"
        yield b""
        return

    wrap(stream, wrapper)
    wrap(AsyncIteratorByteStream.__aiter__, wrapper)

    s = AsyncIteratorByteStream(stream())

    assert b"".join([_ async for _ in s]) == b"hello"
    assert channel == [b"hello", b""]
    with pytest.raises(StreamConsumed):
        b"".join([_ async for _ in s])


@pytest.mark.asyncio
async def test_wrap_async_generator_throw_close():
    channel = []

    async def wrapper(f, args, kwargs):
        nonlocal channel

        channel.append(True)

        __ddgen = f(*args, **kwargs)
        __ddgensend = __ddgen.asend
        try:
            value = await __ddgen.__anext__()
            channel.append(value)
        except StopAsyncIteration:
            return
        while True:
            try:
                tosend = yield value
            except GeneratorExit:
                channel.append("GeneratorExit")
                await __ddgen.aclose()
                raise
            except:  # noqa
                channel.append(sys.exc_info()[0])
                value = await __ddgen.athrow(*sys.exc_info())
                channel.append(value)
            else:
                try:
                    value = await __ddgensend(tosend)
                    channel.append(value)
                except StopAsyncIteration:
                    return

    async def g():
        while True:
            try:
                yield 0
            except ValueError:
                yield 1

    wrap(g, wrapper)
    inspect.isasyncgenfunction(g)

    gen = g()
    inspect.isasyncgen(gen)

    for _ in range(10):
        assert await gen.__anext__() == 0
        assert await gen.athrow(ValueError) == 1

    await gen.aclose()

    assert channel == [True] + [0, ValueError, 1] * 10 + ["GeneratorExit"]



#!/usr/bin/env python3
import logging

from gevent import monkey

from tests.appsec.iast.fixtures.integration.print_str import print_str  # noqa: E402


monkey.patch_all()

logging.basicConfig(level=logging.DEBUG)


def main():
    print_str()


if __name__ == "__main__":
    main()



import sys
from time import sleep

import pytest

from ddtrace.appsec.iast import oce
from ddtrace.appsec.iast._overhead_control_engine import MAX_REQUESTS
from ddtrace.appsec.iast._overhead_control_engine import MAX_VULNERABILITIES_PER_REQUEST
from ddtrace.constants import IAST_CONTEXT_KEY
from ddtrace.internal import _context


def function_with_vulnerabilities_3(tracer):
    with tracer.trace("test_child"):
        import hashlib

        m = hashlib.md5()
        m.update(b"Nobody inspects")
        m.digest()
        sleep(0.3)
    return 1


def function_with_vulnerabilities_2(tracer):
    with tracer.trace("test_child"):
        import hashlib

        m = hashlib.md5()
        m.update(b"Nobody inspects")
        m.digest()
        sleep(0.2)
    return 1


def function_with_vulnerabilities_1(tracer):
    with tracer.trace("test_child"):
        import hashlib

        m = hashlib.md5()
        m.update(b"Nobody inspects")
        m.digest()
        sleep(0.1)
    return 1


@pytest.mark.skipif(sys.version_info < (3, 0, 0), reason="digest works only in Python 3")
def test_oce_max_vulnerabilities_per_request(iast_span_defaults):
    import hashlib

    m = hashlib.md5()
    m.update(b"Nobody inspects")
    m.digest()
    m.digest()
    m.digest()
    m.digest()
    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_defaults)

    assert len(span_report.vulnerabilities) == MAX_VULNERABILITIES_PER_REQUEST


@pytest.mark.skipif(sys.version_info < (3, 0, 0), reason="digest works only in Python 3")
def test_oce_reset_vulnerabilities_report(iast_span_defaults):
    import hashlib

    m = hashlib.md5()
    m.update(b"Nobody inspects")
    m.digest()
    m.digest()
    m.digest()
    oce.vulnerabilities_reset_quota()
    m.digest()

    span_report = _context.get_item(IAST_CONTEXT_KEY, span=iast_span_defaults)

    assert len(span_report.vulnerabilities) == MAX_VULNERABILITIES_PER_REQUEST + 1


def test_oce_max_requests(tracer, iast_span_defaults):
    import threading

    results = []
    num_requests = 5
    total_vulnerabilities = 0

    threads = [threading.Thread(target=function_with_vulnerabilities_1, args=(tracer,)) for _ in range(0, num_requests)]
    for thread in threads:
        thread.start()
    for thread in threads:
        results.append(thread.join())

    spans = tracer.pop()
    for span in spans:
        span_report = _context.get_item(IAST_CONTEXT_KEY, span=span)
        if span_report:
            total_vulnerabilities += len(span_report.vulnerabilities)

    assert len(results) == num_requests
    assert len(spans) == num_requests
    assert total_vulnerabilities == 1


@pytest.mark.skipif(sys.version_info < (3, 0, 0), reason="concurrent.futures exists in Python 3")
def test_oce_max_requests_py3(tracer, iast_span_defaults):
    import concurrent.futures

    results = []
    num_requests = 5
    total_vulnerabilities = 0

    with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
        futures = []
        for _ in range(0, num_requests):
            futures.append(executor.submit(function_with_vulnerabilities_1, tracer))
            futures.append(executor.submit(function_with_vulnerabilities_2, tracer))
            futures.append(executor.submit(function_with_vulnerabilities_3, tracer))

        for future in concurrent.futures.as_completed(futures):
            results.append(future.result())

    spans = tracer.pop()
    for span in spans:
        span_report = _context.get_item(IAST_CONTEXT_KEY, span=span)
        if span_report:
            total_vulnerabilities += len(span_report.vulnerabilities)

    assert len(results) == num_requests * 3
    assert len(spans) == num_requests * 3
    assert total_vulnerabilities == MAX_REQUESTS



import datetime
import sys

import pytest

from ddtrace.internal.utils.time import parse_isoformat


@pytest.mark.parametrize(
    "date_str,expected",
    [
        ("2022-09-01T01:00:00+02:00", datetime.datetime(2022, 8, 31, 23, 0)),
        ("2022-07-31T22:00:00Z", datetime.datetime(2022, 7, 31, 22, 0)),
        ("2022-07-31T22:00:00", None),
        ("2022-07-31", None),
        ("2022-07-3122:00:00", None),
        ("", None),
        ("aa", None),
    ],
)
@pytest.mark.skipif(sys.version_info > (3, 0, 0), reason="Python 2 tests")
def test_parse_isoformat_py2(date_str, expected):
    # type: (str, datetime.datetime) -> None
    result = parse_isoformat(date_str)
    assert result == expected


@pytest.mark.parametrize(
    "date_str,expected",
    [
        ("2022-07-31T22:00:00Z", datetime.datetime(2022, 7, 31, 22, 0)),
        ("2022-07-31T22:00:00", None),
        ("2022-07-31", None),
        ("2022-07-3122:00:00", None),
        ("", None),
        ("aa", None),
    ],
)
@pytest.mark.skipif(sys.version_info < (3, 5, 0) or sys.version_info >= (3, 7, 0), reason="Python 3.5 and 3.6 tests")
def test_parse_isoformat_py36(date_str, expected):
    # type: (str, datetime.datetime) -> None
    result = parse_isoformat(date_str)
    assert result == expected


@pytest.mark.skipif(sys.version_info < (3, 5, 0) or sys.version_info >= (3, 7, 0), reason="Python 3.5 and 3.6 tests")
def test_parse_isoformat_py36_with_timezone():
    # type: () -> None
    result = parse_isoformat("2022-09-01T01:00:00+02:00")
    assert result == datetime.datetime(2022, 8, 31, 23, 0)


@pytest.mark.parametrize(
    "date_str,expected",
    [
        ("2022-07-31T22:00:00Z", datetime.datetime(2022, 7, 31, 22, 0)),
        ("2022-07-31T22:00:00", datetime.datetime(2022, 7, 31, 22, 0)),
        ("2022-07-31", datetime.datetime(2022, 7, 31, 0, 0)),
        ("2022-07-3122:00:00", None),
        ("", None),
        ("aa", None),
    ],
)
@pytest.mark.skipif(sys.version_info < (3, 7, 0), reason="Python 3 tests")
def test_parse_isoformat_py3(date_str, expected):
    # type: (str, datetime.datetime) -> None
    result = parse_isoformat(date_str)
    assert result == expected


@pytest.mark.skipif(sys.version_info < (3, 7, 0), reason="Python 3 tests")
def test_parse_isoformat_py3_with_timezone():
    # type: () -> None
    result = parse_isoformat("2022-09-01T01:00:00+02:00")
    assert result == datetime.datetime(2022, 9, 1, 1, 0, tzinfo=datetime.timezone(datetime.timedelta(seconds=7200)))



"""
The Overhead control engine (OCE) is an element that by design ensures that the overhead does not go over a maximum
limit. It will measure operations being executed in a request and it will deactivate detection
(and therefore reduce the overhead to nearly 0) if a certain threshold is reached.
"""
import os
import threading
from typing import TYPE_CHECKING


if TYPE_CHECKING:  # pragma: no cover
    from typing import Set
    from typing import Tuple
    from typing import Type

MAX_REQUESTS = int(os.environ.get("DD_IAST_MAX_CONCURRENT_REQUESTS", 2))
MAX_VULNERABILITIES_PER_REQUEST = int(os.environ.get("DD_IAST_VULNERABILITIES_PER_REQUEST", 2))


class Operation(object):
    """Common operation related to Overhead Control Engine (OCE). Every vulnerabilities/taint_sinks should inherit
    from this class. OCE instance calls these methods to control the overhead produced in each request.
    """

    _lock = threading.Lock()
    _vulnerability_quota = MAX_VULNERABILITIES_PER_REQUEST
    _reported_vulnerabilities = set()  # type: Set[Tuple[str, int]]

    @classmethod
    def reset(cls):
        cls._vulnerability_quota = MAX_VULNERABILITIES_PER_REQUEST
        cls._reported_vulnerabilities = set()

    @classmethod
    def acquire_quota(cls):
        # type: () -> bool
        cls._lock.acquire()
        result = False
        if cls._vulnerability_quota > 0:
            cls._vulnerability_quota -= 1
            result = True
        cls._lock.release()
        return result

    @classmethod
    def has_quota(cls):
        # type: () -> bool
        cls._lock.acquire()
        result = cls._vulnerability_quota > 0
        cls._lock.release()
        return result

    @classmethod
    def is_not_reported(cls, filename, lineno):
        # type: (str, int) -> bool
        vulnerability_id = (filename, lineno)
        if vulnerability_id in cls._reported_vulnerabilities:
            return False

        cls._reported_vulnerabilities.add(vulnerability_id)
        return True


class OverheadControl(object):
    _request_quota = MAX_REQUESTS
    _enabled = False
    _vulnerabilities = set()  # type: Set[Type[Operation]]

    def acquire_request(self):
        """Block a request's quota at start of the request.

        TODO: Implement sampling request in this method
        """
        if self._request_quota > 0:
            self._request_quota -= 1
            self._enabled = True

    def release_request(self):
        """increment request's quota at end of the request.

        TODO: figure out how to check maximum requests per thread
        """
        if self._request_quota < MAX_REQUESTS:
            self._request_quota += 1
            self._enabled = False
        self.vulnerabilities_reset_quota()

    def register(self, klass):
        # type: (Type[Operation]) -> Type[Operation]
        """Register vulnerabilities/taint_sinks. This set of elements will restart for each request."""
        self._vulnerabilities.add(klass)
        return klass

    @property
    def request_has_quota(self):
        # type: () -> bool
        return self._enabled

    def vulnerabilities_reset_quota(self):
        # type: () -> None
        for k in self._vulnerabilities:
            k.reset()



from typing import Set

import attr


@attr.s(eq=False)
class Evidence(object):
    type = attr.ib(type=str)
    value = attr.ib(type=str, default="")


@attr.s(eq=False)
class Location(object):
    path = attr.ib(type=str)
    line = attr.ib(type=int)


@attr.s(eq=False)
class Vulnerability(object):
    type = attr.ib(type=str)
    evidence = attr.ib(type=Evidence)
    location = attr.ib(type=Location)


@attr.s(eq=False)
class Source(object):
    origin = attr.ib(type=str)
    name = attr.ib(type=str)
    value = attr.ib(type=str)


@attr.s(eq=False)
class IastSpanReporter(object):
    sources = attr.ib(type=Set[Source], factory=set)
    vulnerabilities = attr.ib(type=Set[Vulnerability], factory=set)



import json

import pytest

from ddtrace.constants import APPSEC_JSON
from ddtrace.internal import _context
from ddtrace.internal.compat import urlencode
from ddtrace.internal.constants import APPSEC_BLOCKED_RESPONSE_JSON
from tests.appsec.test_processor import RULES_GOOD_PATH
from tests.appsec.test_processor import _BLOCKED_IP
from tests.appsec.test_telemety import _assert_distributions_metrics
from tests.appsec.test_telemety import _assert_generate_metrics
from tests.appsec.test_telemety import mock_telemetry_metrics_writer  # noqa: F401
from tests.contrib.flask import BaseFlaskTestCase
from tests.utils import override_env
from tests.utils import override_global_config


class FlaskAppSecTestCase(BaseFlaskTestCase):
    @pytest.fixture(autouse=True)
    def inject_fixtures(self, mock_telemetry_metrics_writer):  # noqa: F811
        self.mock_telemetry_metrics_writer = mock_telemetry_metrics_writer

    def _aux_appsec_prepare_tracer(self, appsec_enabled=True):
        self.tracer._appsec_enabled = appsec_enabled
        # Hack: need to pass an argument to configure so that the processors are recreated
        self.tracer.configure(api_version="v0.4")

    def test_telemetry_metrics_block(self):
        with override_global_config(dict(_appsec_enabled=True, _telemetry_metrics_enabled=True)), override_env(
            dict(DD_APPSEC_RULES=RULES_GOOD_PATH)
        ):
            self._aux_appsec_prepare_tracer()
            resp = self.client.get("/", headers={"X-Real-Ip": _BLOCKED_IP})
            assert resp.status_code == 403
            if hasattr(resp, "text"):
                assert resp.text == APPSEC_BLOCKED_RESPONSE_JSON

        _assert_generate_metrics(
            self.mock_telemetry_metrics_writer._namespace._metrics_data, is_rule_triggered=True, is_blocked_request=True
        )
        _assert_distributions_metrics(
            self.mock_telemetry_metrics_writer._namespace._metrics_data, is_rule_triggered=True, is_blocked_request=True
        )

    def test_telemetry_metrics_attack(self):
        with override_global_config(dict(_appsec_enabled=True, _telemetry_metrics_enabled=True)):
            self._aux_appsec_prepare_tracer()
            payload = urlencode({"attack": "1' or '1' = '1'"})
            self.client.post("/", data=payload, content_type="application/x-www-form-urlencoded")
            root_span = self.pop_spans()[0]
            query = dict(_context.get_item("http.request.body", span=root_span))
            assert "triggers" in json.loads(root_span.get_tag(APPSEC_JSON))
            assert query == {"attack": "1' or '1' = '1'"}

        _assert_generate_metrics(
            self.mock_telemetry_metrics_writer._namespace._metrics_data,
            is_rule_triggered=True,
            is_blocked_request=False,
        )
        _assert_distributions_metrics(
            self.mock_telemetry_metrics_writer._namespace._metrics_data,
            is_rule_triggered=True,
            is_blocked_request=False,
        )
